@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2020 GraphicsMagick Group
+% Copyright (C) 2003-2021 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -44,47 +44,60 @@
 #include "magick/magick.h"
 #include "magick/monitor.h"
 #include "magick/profile.h"
+#include "magick/resource.h"
 #include "magick/utility.h"
 #if defined(HasJP2)
-#if !defined(uchar)
-#define uchar  unsigned char
-#endif
-#if !defined(ushort)
-#define ushort  unsigned short
-#endif
-#if !defined(uint)
-#define uint  unsigned int
-#endif
-#if !defined(longlong)
-#define longlong  long long
-#endif
-#if !defined(ulonglong)
-#define ulonglong  unsigned long long
-#endif
+#  if !defined(uchar)
+#    define uchar  unsigned char
+#  endif
+#  if !defined(ushort)
+#    define ushort  unsigned short
+#  endif
+#  if !defined(uint)
+#    define uint  unsigned int
+#  endif
+#  if !defined(longlong)
+#    define longlong  long long
+#  endif
+#  if !defined(ulonglong)
+#    define ulonglong  unsigned long long
+#  endif
+
+#  ifdef __VMS
+#    define JAS_VERSION 1.700.0
+#    define PACKAGE jasper
+#    define VERSION 1.700.0
+#  endif /* ifdef __VMS */
+#  undef PACKAGE_NAME
+#  undef PACKAGE_STRING
+#  undef PACKAGE_TARNAME
+#  undef PACKAGE_VERSION
+#  include "jasper/jasper.h"
+#  undef PACKAGE_NAME
+#  undef PACKAGE_STRING
+#  undef PACKAGE_TARNAME
+#  undef PACKAGE_VERSION
+/*
+  Old JasPer uses non-persistent '!defined(EXCLUDE_FOO_SUPPORT)' and
+  modern JasPer uses persistent 'if defined(JAS_INCLUDE_FOO_CODEC)'
+  in jas_image.h
+*/
+#  if defined(EXCLUDE_JP2_SUPPORT)
+#    undef HAVE_JP2_DECODE
+#  endif
+#  if defined(EXCLUDE_JPC_SUPPORT)
+#    undef HAVE_JPC_DECODE
+#  endif
+#  if defined(EXCLUDE_PGX_SUPPORT)
+#    undef HAVE_PGX_DECODE
+#  endif
 
-#ifdef __VMS
-#define JAS_VERSION 1.700.0
-#define PACKAGE jasper
-#define VERSION 1.700.0
-#endif
-#undef PACKAGE_NAME
-#undef PACKAGE_STRING
-#undef PACKAGE_TARNAME
-#undef PACKAGE_VERSION
-#include "jasper/jasper.h"
-#undef PACKAGE_NAME
-#undef PACKAGE_STRING
-#undef PACKAGE_TARNAME
-#undef PACKAGE_VERSION
-#endif
 
 /*
   Forward declarations.
 */
-#if defined(HasJP2)
 static unsigned int
   WriteJP2Image(const ImageInfo *,Image *);
-#endif
 
 static MagickBool jasper_initialized=MagickFalse;
 static const char jasper_options[][11] =
@@ -265,7 +278,6 @@ static unsigned int IsPGX(const unsigned char *magick,const size_t length)
 %    o exception: return any errors or warnings in this structure.
 %
 */
-#if defined(HasJP2)
 
 typedef struct _StreamManager
 {
@@ -294,14 +306,14 @@ typedef struct _StreamManager
   We have yet to find a useful way to determine the version of the
   JasPer library using the C pre-processor.
  */
-#if !defined(MAGICK_JP2_NEW_STREAM_INTERFACE)
-#define MAGICK_JP2_NEW_STREAM_INTERFACE 0
-#endif /* if !defined(MAGICK_JP2_NEW_STREAM_INTERFACE) */
+#if !defined(MAGICK_JP2_OLD_STREAM_INTERFACE)
+#  define MAGICK_JP2_OLD_STREAM_INTERFACE 0
+#endif /* if !defined(MAGICK_JP2_OLD_STREAM_INTERFACE) */
 
-#if MAGICK_JP2_NEW_STREAM_INTERFACE
-static int BlobRead(jas_stream_obj_t *object,char *buffer,unsigned length)
-#else
+#if MAGICK_JP2_OLD_STREAM_INTERFACE
 static int BlobRead(jas_stream_obj_t *object,char *buffer,const int length)
+#else
+static int BlobRead(jas_stream_obj_t *object,char *buffer,unsigned length)
 #endif
 {
   size_t
@@ -314,10 +326,10 @@ static int BlobRead(jas_stream_obj_t *object,char *buffer,const int length)
   return ((int) count);
 }
 
-#if MAGICK_JP2_NEW_STREAM_INTERFACE
-static int BlobWrite(jas_stream_obj_t *object,const char *buffer,unsigned length)
-#else
+#if MAGICK_JP2_OLD_STREAM_INTERFACE
 static int BlobWrite(jas_stream_obj_t *object,char *buffer,const int length)
+#else
+static int BlobWrite(jas_stream_obj_t *object,const char *buffer,unsigned length)
 #endif
 {
   size_t
@@ -344,7 +356,7 @@ static int BlobClose(jas_stream_obj_t *object)
     *source = (StreamManager *) object;
 
   CloseBlob(source->image);
-  MagickFreeMemory(source);
+  jas_free(source);
   return (0);
 }
 
@@ -357,15 +369,15 @@ static jas_stream_t *JP2StreamManager(jas_stream_ops_t *stream_ops, Image *image
   StreamManager
     *source;
 
-  stream=MagickAllocateMemory(jas_stream_t *,sizeof(jas_stream_t));
+  stream=(jas_stream_t *) jas_malloc(sizeof(jas_stream_t));
   if (stream == (jas_stream_t *) NULL)
     return((jas_stream_t *) NULL);
   (void) memset(stream,0,sizeof(jas_stream_t));
   stream->rwlimit_=(-1);
-  stream->obj_=MagickAllocateMemory(jas_stream_obj_t *,sizeof(StreamManager));
+  stream->obj_=(jas_stream_obj_t *) jas_malloc(sizeof(StreamManager));
   if (stream->obj_ == (jas_stream_obj_t *) NULL)
     {
-      MagickFreeMemory(stream);
+      jas_free(stream);
       return((jas_stream_t *) NULL);
     }
   stream->ops_=stream_ops;
@@ -393,8 +405,92 @@ static jas_stream_t *JP2StreamManager(jas_stream_ops_t *stream_ops, Image *image
   ThrowReaderException(code_,reason_,image_); \
 }
 
+/*
+  Initialize Jasper
+*/
+#if HAVE_JAS_INITIALIZE
+static void *alloc_rlm(struct jas_allocator_s *allocator, size_t size)
+{
+  (void) allocator;
+  /* JasPer expects its allocator to return non-null for zero size */
+  return _MagickAllocateResourceLimitedMemory(size == 0 ? 1 : size);
+}
+static void free_rlm(struct jas_allocator_s *allocator, void *pointer)
+{
+  (void) allocator;
+  _MagickFreeResourceLimitedMemory(pointer);
+}
+static void *realloc_rlm(struct jas_allocator_s *allocator, void *pointer,
+                         size_t new_size)
+{
+  (void) allocator;
+  return _MagickReallocateResourceLimitedMemory(pointer,1,new_size,0);
+}
+#endif /* if HAVE_JAS_INITIALIZE */
+static void initialize_jasper(void)
+{
+  if (!jasper_initialized)
+    {
+#if HAVE_JAS_INITIALIZE
+      {
+        /* static jas_std_allocator_t allocator; */
+        static jas_allocator_t allocator;
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Initializing JasPer...");
+        jas_conf_clear();
+        /* Use our own resource-limited memory allocation functions */
+        allocator.cleanup = 0;
+        allocator.alloc = alloc_rlm;
+        allocator.free = free_rlm;
+        allocator.realloc = realloc_rlm;
+        /* jas_std_allocator_init(&allocator); */ /* Uses JasPer allocators */
+        jas_conf_set_allocator(&allocator); /* Assigns jas_allocator_t to jas_conf.allocator in library */
+        /* jas_conf_set_debug_level(cmdopts->debug); */
+        /* jas_conf_set_max_mem(cmdopts->max_mem); */ /* Only pertains to JasPer allocators */
+
+        if (jas_initialize() == 0)
+          {
+            jasper_initialized=MagickTrue;
+          }
+      }
+#else
+      {
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Initializing JasPer...");
+        if (jas_init() == 0)
+          {
+            jasper_initialized=MagickTrue;
+          }
+      }
+#endif  /* HAVE_JAS_INITIALIZE */
+
+      if (!jasper_initialized)
+        {
+          (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                "Failed to initialize JasPer!");
+        }
+    }
+}
+
+
+/*
+  Cleanup Jasper
+*/
+static void cleanup_jasper(void)
+{
+  if (jasper_initialized)
+    {
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                            "Destroying JasPer...");
+#if HAVE_JAS_INITIALIZE
+      jas_cleanup();
+#endif /* if HAVE_JAS_INITIALIZE */
+      jasper_initialized=MagickFalse;
+    }
+}
+
 static Image *ReadJP2Image(const ImageInfo *image_info,
-  ExceptionInfo *exception)
+                           ExceptionInfo *exception)
 {
   Image
     *image;
@@ -426,6 +522,9 @@ static Image *ReadJP2Image(const ImageInfo *image_info,
   register PixelPacket
     *q;
 
+  magick_off_t
+    pos;
+
   int
     component,
     components[4],
@@ -434,19 +533,21 @@ static Image *ReadJP2Image(const ImageInfo *image_info,
   Quantum
     *channel_lut[4];
 
+  char
+    options[MaxTextExtent];
+
   unsigned int
     status;
 
+  MagickBool
+    jp2_hdr,
+    jpc_hdr,
+    pgx_hdr;
+
   /*
     Initialize Jasper
   */
-#if defined(HasJP2)
-  if (!jasper_initialized)
-    {
-      jas_init();
-      jasper_initialized=MagickTrue;
-    }
-#endif
+  initialize_jasper();
 
   /*
     Open image file.
@@ -456,18 +557,140 @@ static Image *ReadJP2Image(const ImageInfo *image_info,
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickSignature);
   (void) memset(channel_lut,0,sizeof(channel_lut));
+  options[0]='\0';
   image=AllocateImage(image_info);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == False)
     ThrowReaderException(FileOpenError,UnableToOpenFile,image);
 
+  (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                        "Requested format is \"%s\"",
+                        image_info->magick);
+
+  /*
+    Get the header and auto-detect apparent format.
+  */
+  {
+    size_t
+      magick_length;
+
+    unsigned char
+      magick[16];
+
+    const MagickInfo
+      *magick_info;
+
+    /* Get current seek position (normally 0) */
+    pos=TellBlob(image);
+
+    /* Read header */
+    if ((magick_length=ReadBlob(image,sizeof(magick),magick)) != sizeof(magick))
+      {
+        ThrowReaderException(CorruptImageError,UnexpectedEndOfFile,
+                             image);
+      }
+
+    /* Restore seek position */
+    if (SeekBlob(image,pos,SEEK_SET) != pos)
+      {
+        ThrowReaderException(BlobError,UnableToSeekToOffset,image);
+      }
+
+    /* Inspect header to see what it might actually be */
+    jp2_hdr=IsJP2(magick,sizeof(magick));
+    jpc_hdr=IsJPC(magick,sizeof(magick));
+    pgx_hdr=IsPGX(magick,sizeof(magick));
+
+    /*
+      If input format was previously auto-detected or specified, then
+      assure that header matches what was specified.
+    */
+    if (((LocaleCompare(image_info->magick,"JP2") == 0) && !jp2_hdr) ||
+        (((LocaleCompare(image_info->magick,"JPC") == 0) ||
+          (LocaleCompare(image_info->magick,"J2C") == 0)) && !jpc_hdr) ||
+        ((LocaleCompare(image_info->magick,"PGX") == 0) && !pgx_hdr))
+      {
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Not a \"%s\" file!", image_info->magick);
+        ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+      }
+
+    /*
+      Throw exception if header is not one we expect.
+    */
+    if (!jp2_hdr && !jpc_hdr && !pgx_hdr)
+      {
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Header is not a supported type for this coder");
+        ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+      }
+
+    /*
+      Check if we are allowed to decode this format.
+    */
+    if (((magick_info = GetMagickInfo(image_info->magick,exception)) ==
+         (const MagickInfo *) NULL) ||
+        (magick_info->decoder == (DecoderHandler) NULL))
+      {
+        ThrowReaderException(DelegateError,UnableToDecodeImageFile,image);
+      }
+  }
+
   /*
     Obtain a JP2 Stream.
   */
   jp2_stream=JP2StreamManager(&StreamOperators, image);
   if (jp2_stream == (jas_stream_t *) NULL)
     ThrowReaderException(DelegateError,UnableToManageJP2Stream,image);
-  jp2_image=jas_image_decode(jp2_stream,-1,0);
+
+  {
+    /* Pass options argument which specifies "max_samples" to cap memory usage. */
+    const magick_uint64_t memory_limit = (magick_uint64_t) GetMagickResourceLimit(MemoryResource);
+    const magick_uint64_t max_samples = memory_limit/4U * sizeof(magick_uint32_t);
+
+    FormatString(options,"max_samples=%" MAGICK_UINT64_F "u", max_samples);
+    (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                          "JP2 options = \"%s\"", options);
+  }
+
+  /* JP2, JPC, PGX (J2C is an alias for JPC) */
+#if HAVE_JP2_DECODE
+  /* JPEG-2000 JP2 File Format Syntax */
+  if (jp2_hdr)
+    {
+      /* jas_image_t *jp2_decode(jas_stream_t *in, const char *optstr); */
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                            "Decoding JP2...");
+      jp2_image=jp2_decode(jp2_stream,options);
+    }
+#endif /* if HAVE_JP2_DECODE */
+#if HAVE_JPC_DECODE
+  /* JPEG-2000 Code Stream Syntax */
+  if (jpc_hdr)
+    {
+      /* jas_image_t *jpc_decode(jas_stream_t *in, const char *optstr); */
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                            "Decoding JPC...");
+      jp2_image=jpc_decode(jp2_stream,options);
+    }
+#endif /* if HAVE_JPC_DECODE */
+#if HAVE_PGX_DECODE
+  /* JPEG-2000 VM Format */
+  if (pgx_hdr)
+    {
+      /* jas_image_t *pgx_decode(jas_stream_t *in, const char *optstr); */
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                            "Decoding PGX...");
+      jp2_image=pgx_decode(jp2_stream,options);
+    }
+#endif /* if HAVE_PGX_DECODE */
+
+  /*
+    Using jas_image_decode() makes us subject to Jasper's own format
+    determination, which may include file formats we don't want to
+    support via Jasper.
+  */
+  /* jp2_image=jas_image_decode(jp2_stream,-1,options); */
   if (jp2_image == (jas_image_t *) NULL)
     ThrowJP2ReaderException(DelegateError,UnableToDecodeImageFile,image);
 
@@ -665,7 +888,7 @@ static Image *ReadJP2Image(const ImageInfo *image_info,
           for (x=0; x < (long) image->columns; x++)
             q[x].blue=(channel_lut[2])[jas_matrix_getv(pixels,x)];
 
-            /* Opacity */
+          /* Opacity */
           if (number_components > 3)
             {
               (void) jas_image_readcmpt(jp2_image,(short) components[3],0,
@@ -743,7 +966,7 @@ static Image *ReadJP2Image(const ImageInfo *image_info,
   StopTimer(&image->timer);
   return(image);
 }
-#endif
+#endif /* if defined(HasJP2) */
 
 /*
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -770,64 +993,76 @@ static Image *ReadJP2Image(const ImageInfo *image_info,
 */
 ModuleExport void RegisterJP2Image(void)
 {
+#if defined(HasJP2)
+  static char
+    version[16];
+
   MagickInfo
     *entry;
 
+  (void) strlcpy(version,"JasPer ",sizeof(version));
+  (void) strlcat(version,jas_getversion(),sizeof(version));
+
+#if HAVE_JPC_DECODE
   entry=SetMagickInfo("J2C");
   entry->description="JPEG-2000 Code Stream Syntax";
+  entry->version=version;
   entry->module="JP2";
   entry->magick=(MagickHandler) IsJPC;
   entry->adjoin=False;
   entry->seekable_stream=True;
   entry->thread_support=False;
-#if defined(HasJP2)
   entry->decoder=(DecoderHandler) ReadJP2Image;
   entry->encoder=(EncoderHandler) WriteJP2Image;
-#endif
   entry->coder_class=StableCoderClass;
   (void) RegisterMagickInfo(entry);
+#endif /* if HAVE_JPC_DECODE */
 
+#if HAVE_JP2_DECODE
   entry=SetMagickInfo("JP2");
   entry->description="JPEG-2000 JP2 File Format Syntax";
+  entry->version=version;
   entry->module="JP2";
   entry->magick=(MagickHandler) IsJP2;
   entry->adjoin=False;
   entry->seekable_stream=True;
   entry->thread_support=False;
-#if defined(HasJP2)
   entry->decoder=(DecoderHandler) ReadJP2Image;
   entry->encoder=(EncoderHandler) WriteJP2Image;
-#endif
   entry->coder_class=StableCoderClass;
   (void) RegisterMagickInfo(entry);
+#endif /* HAVE_JP2_DECODE */
 
+#if HAVE_JPC_DECODE
   entry=SetMagickInfo("JPC");
   entry->description="JPEG-2000 Code Stream Syntax";
+  entry->version=version;
   entry->module="JP2";
   entry->magick=(MagickHandler) IsJPC;
   entry->adjoin=False;
   entry->seekable_stream=True;
   entry->thread_support=False;
-#if defined(HasJP2)
   entry->decoder=(DecoderHandler) ReadJP2Image;
   entry->encoder=(EncoderHandler) WriteJP2Image;
-#endif
   entry->coder_class=StableCoderClass;
   (void) RegisterMagickInfo(entry);
+#endif /* if HAVE_JPC_DECODE */
 
+#if HAVE_PGX_DECODE
   entry=SetMagickInfo("PGX");
   entry->description="JPEG-2000 VM Format";
+  entry->version=version;
   entry->module="JP2";
   entry->magick=(MagickHandler) IsPGX;
   entry->adjoin=False;
   entry->seekable_stream=True;
   entry->thread_support=False;
-#if defined(HasJP2)
   entry->decoder=(DecoderHandler) ReadJP2Image;
   entry->encoder=(EncoderHandler) WriteJP2Image;
-#endif
-  entry->coder_class=StableCoderClass;
+  entry->coder_class=UnstableCoderClass;
   (void) RegisterMagickInfo(entry);
+#endif /* if HAVE_PGX_DECODE */
+#endif /* if defined(HasJP2) */
 }
 
 /*
@@ -842,7 +1077,7 @@ ModuleExport void RegisterJP2Image(void)
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
 %  Method UnregisterJP2Image removes format registrations made by the
-%  PNG module from the list of supported formats.
+%  JP2 module from the list of supported formats.
 %
 %  The format of the UnregisterJP2Image method is:
 %
@@ -851,21 +1086,17 @@ ModuleExport void RegisterJP2Image(void)
 */
 ModuleExport void UnregisterJP2Image(void)
 {
+#if defined(HasJP2)
   (void) UnregisterMagickInfo("PGX");
   (void) UnregisterMagickInfo("JPC");
   (void) UnregisterMagickInfo("JP2");
   (void) UnregisterMagickInfo("J2C");
 
-#if defined(HasJP2)
   /*
     Cleanup Jasper
   */
-  if (jasper_initialized)
-    {
-      jas_cleanup();
-      jasper_initialized=MagickFalse;
-    }
-#endif
+  cleanup_jasper();
+#endif /* if defined(HasJP2) */
 }
 
 #if defined(HasJP2)
@@ -958,13 +1189,7 @@ WriteJP2Image(const ImageInfo *image_info,Image *image)
   /*
     Initialize Jasper
   */
-#if defined(HasJP2)
-  if (!jasper_initialized)
-    {
-      jas_init();
-      jasper_initialized=MagickTrue;
-    }
-#endif
+  initialize_jasper();
 
   /*
     Open image file.
@@ -1252,4 +1477,4 @@ WriteJP2Image(const ImageInfo *image_info,Image *image)
     ThrowWriterException(DelegateError,UnableToEncodeImageFile,image);
   return(True);
 }
-#endif
+#endif /* if defined(HasJP2) */
